/******************************************************************************

Copyright (c) 2020 The Linux Foundation. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above
          copyright notice, this list of conditions and the following
          disclaimer in the documentation and/or other materials provided
          with the distribution.
        * Neither the name of The Linux Foundation nor the names of its
          contributors may be used to endorse or promote products derived
          from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT
ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************/

/*===========================================================================
                              INCLUDE FILES
===========================================================================*/

#include "mmdata_qrtr.h"

extern FILE* log_file;

qrtr_client_info mmdata_fmt_info;
qrtr_client_info mmdata_port_info;

void print_qrtr_msg
(
  const unsigned char  *msg,
  int                  msg_len
)
{
  int i, j;
  char buff[MSG_NUM_CHARS_PER_BYTE * MSG_NUM_BYTES_PER_LINE + 1] = "";
  unsigned char upper_half;
  unsigned char lower_half;
  const unsigned char *data = msg;

  if (NULL == msg)
  {
    return;
  }

  for (i = 1, j = 0; i <= msg_len; ++i, ++data)
  {
    upper_half = (*data) >> 4;
    lower_half = (*data) & 0x0F;
    buff[j++] = GET_HEX_CHAR(upper_half);
    buff[j++] = GET_HEX_CHAR(lower_half);
    buff[j++] = ' ';

    if (i % MSG_NUM_BYTES_PER_LINE == 0)
    {
      buff[j] = '\0';
      MMDATA_INFO ("%s\n", buff);
      j = 0;
    }
  }

  /* Print any remaining data */
  if (j > 0)
  {
    buff[j] = '\0';
    MMDATA_INFO ("%s\n", buff);
  }
}

int qrtr_client_init
(
  qrtr_client_info  *client_info,
  int             timeout
)
{
  int ret=MMDATA_FAILURE;
  int rc= 0;
  unsigned int qrtr_cmd_type;
  struct qrtr_ctrl_pkt qrtr_lookup_resp;

  memset(&qrtr_lookup_resp, 0, sizeof(qrtr_lookup_resp));

  do
  {
    /* Open QRTR Lookup Socket */
    client_info->lookup_sock = qrtr_open(QRTR_CLIENT_PORT);

    if (client_info->lookup_sock < 0)
    {
      MMDATA_ERRNO("%s qrtr_open", __func__);
      break;
    }

    qrtr_cmd_type = QRTR_TYPE_NEW_LOOKUP;
    if (MMDATA_SUCCESS != qrtr_new_lookup( client_info->lookup_sock,
                                        client_info->service,
                                        client_info->version,
                                        client_info->instance))
    {
      MMDATA_ERR("%s qrtr_new_lookup failed lookup_sock=%d\n",__func__, client_info->lookup_sock);
      break;
    }

    MMDATA_INFO("%s qrtr_new_lookup success lookup_sock=%d\n",
                  __func__, client_info->lookup_sock);

    /* Poll on Lookup Socket for response */
    qrtr_poll(client_info->lookup_sock, timeout);

    /* Get response from QRTR */
    rc = qrtr_recv(client_info->lookup_sock, &qrtr_lookup_resp, sizeof(qrtr_lookup_resp));
    if (rc < 0 || rc < sizeof(qrtr_lookup_resp))
    {
      MMDATA_ERR("%s qrtr_recv Invalid Resp lookup_sock=%d rc=%d\n", __func__, client_info->lookup_sock, rc);
      break;
    }

    /* Process Lookup Response */
    qrtr_cmd_type = le32_to_cpu(qrtr_lookup_resp.cmd);

    if (qrtr_cmd_type != QRTR_TYPE_NEW_SERVER)
    {
      MMDATA_ERR("%s qrtr_recv Wrong CMD type: %u\n", __func__, qrtr_cmd_type);
      break;
    }

    if ( !qrtr_lookup_resp.server.service && !qrtr_lookup_resp.server.instance &&
       !qrtr_lookup_resp.server.node && !qrtr_lookup_resp.server.port)
    {
      MMDATA_ERR("%s qrtr_recv Invalid Resp\n", __func__);
      break;
    }

    client_info->service   = le32_to_cpu(qrtr_lookup_resp.server.service);
    client_info->version   = le32_to_cpu(qrtr_lookup_resp.server.instance) & 0xff;
    client_info->instance  = le32_to_cpu(qrtr_lookup_resp.server.instance) >> 8;
    client_info->node      = le32_to_cpu(qrtr_lookup_resp.server.node);
    client_info->port      = le32_to_cpu(qrtr_lookup_resp.server.port);

    MMDATA_INFO("  Service Version Instance Node  Port\n");
    MMDATA_INFO("%9d %7d %8d %4d %5d\n",
           client_info->service,
           client_info->version,
           client_info->instance,
           client_info->node,
           client_info->port);
      ret = MMDATA_SUCCESS;
    }while(0);

  return ret;
}

int
qrtr_client_release
(
  qrtr_client_info  *client_info
)
{

  if (client_info->lookup_sock)
  {
    qrtr_close(client_info->lookup_sock);
    client_info->lookup_sock=0;
  }

  return MMDATA_SUCCESS;
}

int qrtr_send_req
(
  qrtr_client_info *client_info,
  void * rdata,
  int  rlen,
  int  timeout_msecs,
  int  close_sock
)
{
  int rc = 0;
  int ret = MMDATA_FAILURE;
  uint32_t recv_node = 0;
  uint32_t recv_port = 0;
  unsigned char *msg_ptr = NULL;
  struct qrtr_packet open_req_buf;
  struct qrtr_packet open_resp_buf;

  do
  {

    if ( NULL == client_info )
    {
      MMDATA_ERR("%s Invalid Parameter\n", __func__);
      break;
    }

    if (client_info->data_sock <= 0)
    {
      client_info->data_sock = qrtr_open(QRTR_CLIENT_PORT);
      MMDATA_INFO("%s qrtr_open data_sock=%d\n", __func__, client_info->data_sock);
    }

    if (client_info->data_sock < 0)
    {
      MMDATA_ERRNO("%s qrtr_open errno %d ", __func__,errno);
      break;
    }

    open_req_buf.node       = client_info->node;
    open_req_buf.port       = client_info->port;
    open_req_buf.service    = client_info->service;
    open_req_buf.instance   = client_info->instance;
    open_req_buf.version    = client_info->version;

    open_req_buf.data = (void *) mmdata_malloc(rlen+1);

    if (NULL == open_req_buf.data)
    {
      printf("%s failed to allocate memory for open_req_buf\n", __func__);
      break;
    }

    memset(open_req_buf.data,0x00,rlen+1);
    memcpy(open_req_buf.data,rdata,rlen);
    open_req_buf.data_len = rlen;

    MMDATA_ERR("%s", "TX  send msg:\n");

    if (MMDATA_SUCCESS != qrtr_sendto(client_info->data_sock,
                                   client_info->node,
                                   client_info->port,
                                   open_req_buf.data,
                                   open_req_buf.data_len))
    {
      MMDATA_ERR("%s qrtr_sendto FAILED:\n", __func__);
      break;
    }

    /* Poll on Data Socket for response */
    MMDATA_INFO("%s qrtr_poll data_sock=%d\n", __func__, client_info->data_sock);
    qrtr_poll(client_info->data_sock, timeout_msecs);

    /* Allocate resp buffer with max resp size */
    open_resp_buf.node      = client_info->node;
    open_resp_buf.port      = client_info->port;
    open_resp_buf.service   = client_info->service;
    open_resp_buf.instance  = client_info->instance;
    open_resp_buf.version   = client_info->version;

    open_resp_buf.data = (void *) mmdata_malloc(256);
    if (NULL == open_resp_buf.data)
    {
      MMDATA_ERR("%s failed to allocate memory for open_resp_buf\n", __func__);
      break;
    }

    rc = qrtr_recvfrom(client_info->data_sock,
                   open_resp_buf.data,
                   256,
                   &recv_node,
                   &recv_port);
    if (rc < 0 || recv_node != client_info->node || recv_port != client_info->port)
    {
      MMDATA_ERR("%s qrtr_recvfrom FAILED rc=%d, recv_node=%d, recv_port=%d\n",
             __func__, rc, recv_node, recv_port);
      break;
    }
    else
    {
      open_resp_buf.data_len = rc;
      MMDATA_INFO("%s qrtr_recvfrom SUCCESS rc=%d\n", __func__, rc);
    }

    MMDATA_INFO("%s", "RX recvd msg:\n");
    msg_ptr = (unsigned char *)open_resp_buf.data;
    print_qrtr_msg(msg_ptr,(int) open_resp_buf.data_len);
    msg_ptr = NULL;

    ret = MMDATA_SUCCESS;
  }while(0);

  /* Cleanup before exit */
  if (open_req_buf.data != NULL)
  {
    mmdata_free(open_req_buf.data);
  }

  if (open_resp_buf.data != NULL)
  {
    mmdata_free(open_resp_buf.data);
  }

  if (TRUE == close_sock && client_info->data_sock)
  {
    MMDATA_INFO("%s qrtr_close data_sock=%d\n", __func__, client_info->data_sock);
    qrtr_close(client_info->data_sock);
    client_info->data_sock = 0;
  }

  return ret;
}

int mmdata_port_cfg(void)
{

  int ret = MMDATA_FAILURE;
  unsigned char port_cfg_req[]={ 0x00, 0x03, 0x00, 0x20, 0x00, 0x14 ,0x00, 0x11,
                            0x11, 0x00, 0x01, 0x04, 0x00, 0x00, 0x00, 0x01,
                            0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x09,
                            0x00,0x00,0x00};

  do
  {
    memset(&mmdata_port_info, 0, sizeof(qrtr_client_info));
    mmdata_port_info.service = (0x2F);
    mmdata_port_info.version = (0x01);
    mmdata_port_info.instance = (0x00);

    if (MMDATA_SUCCESS != qrtr_client_init(&mmdata_port_info, MMDATA_QRTR_TIMEOUT))
    {
      MMDATA_ERR("%s qrtr_client_init FAILED\n", __func__);
      break;
    }

    ret = qrtr_send_req(&mmdata_port_info,
                     (void *)port_cfg_req,
                     sizeof(port_cfg_req),
                     MMDATA_QRTR_TIMEOUT,
                     TRUE);
    if (MMDATA_SUCCESS != ret)
    {
      MMDATA_ERR("%s send open_req failed, err %d\n", __func__, ret);
      break;
    }

    ret = MMDATA_SUCCESS;
    MMDATA_INFO("%s : PORT CFG  SUCCESS ret=%d\n", __func__, ret);

  }while (0);

  return ret;

}

int mmdata_fmt_cfg(void)
{
  int ret = MMDATA_FAILURE;
  unsigned char fmt_cfg_req[]= { 0x00, 0x03, 0x00, 0x20, 0x00, 0x2E, 0x00, 0x11,
                            0x04, 0x00, 0x02, 0x00, 0x00, 0x00, 0x12, 0x04,
                            0x00, 0x05, 0x00, 0x00, 0x00, 0x13, 0x04, 0x00,
                            0x05, 0x00, 0x00, 0x00, 0x15, 0x04, 0x00, 0x01,
                            0x00, 0x00, 0x00, 0x16, 0x04, 0x00, 0xE0, 0x05,
                            0x00, 0x00, 0x17, 0x08, 0x00, 0x04, 0x00, 0x00,
                            0x00, 0x01, 0x00, 0x00, 0x00 };
  do
  {
    memset(&mmdata_fmt_info, 0, sizeof(qrtr_client_info));
    mmdata_fmt_info.service = (0x1A);
    mmdata_fmt_info.version = (0x01);
    mmdata_fmt_info.instance = (0x00);


    if (MMDATA_SUCCESS != qrtr_client_init(&mmdata_fmt_info, MMDATA_QRTR_TIMEOUT))
    {
      MMDATA_ERR("%s qrtr_client_init FAILED\n", __func__);
      break;
    }

    ret = qrtr_send_req(&mmdata_fmt_info,
                     (void *)fmt_cfg_req,
                     sizeof(fmt_cfg_req),
                     MMDATA_QRTR_TIMEOUT,
                     TRUE);

    if (MMDATA_SUCCESS != ret)
    {
      MMDATA_ERR("%s send format failed, err %d\n", __func__, ret);
      break;
    }

    ret = MMDATA_SUCCESS;
    MMDATA_INFO("%s  Format CFG Success ret=%d\n", __func__, ret);

  }while (0);

  return ret;

}
